﻿using System;
using System.Collections.Generic;
using System.IO;
using UnityEditor;
using UnityEngine;
using UnityEngine.SceneManagement;

namespace VIRNECT
{
    [CustomEditor(typeof(TrackManager))]
    public class TrackManagerEditor : Editor
    {
        // References to serialized fields
        SerializedProperty camera;
        SerializedProperty cameraID;
        SerializedProperty imageSource;
        SerializedProperty fileSquencePath;
        SerializedProperty fileSequenceCalibrationPath;

        string[] imageSourceNames;
        int imageSourceDropdownIndex = 0;

        /// <summary>
        /// Prepare connection to serializedObject
        /// </summary>
        void OnEnable()
        {
            // Load data reference to entity
            camera = serializedObject.FindProperty("ARCamera");
            cameraID = serializedObject.FindProperty("cameraID");
            imageSource = serializedObject.FindProperty("imageSource");
            fileSquencePath = serializedObject.FindProperty("fileSequencePath");
            fileSequenceCalibrationPath = serializedObject.FindProperty("fileSequenceCalibrationPath");

            imageSourceNames = Enum.GetNames(typeof(TrackManager.ImageSources));
            imageSourceDropdownIndex = imageSource.intValue;
        }


        /// <summary>
        /// Custom UI, alerts the user when no camera is attached, or the attached camera has no TrackCamera script attached
        /// </summary>
        public override void OnInspectorGUI()
        {
            serializedObject.Update();
            GUILayout.BeginVertical();

            EditorGUILayout.LabelField("Scene References");

            // Camera reference
            EditorGUILayout.PropertyField(camera, new GUIContent("Scene Camera"));

            // Alert if empty
            if (camera.objectReferenceValue == null)
                EditorGUILayout.HelpBox("Please reference the main scene camera.", MessageType.Error, false);
            else
            {
                // Alert if no TrackCamera component is attached
                Camera c = (Camera)camera.objectReferenceValue;
                if (c.GetComponent<TrackCamera>() == null)
                {
                    // Alert when no script attached
                    EditorGUILayout.HelpBox("The referenced camera has no TrackCamera script attached!", MessageType.Error, false);
                    if (GUILayout.Button("Add TrackCamera script to \"" + c.gameObject.name + "\"", EditorStyles.miniButton))
                    {
                        c.gameObject.AddComponent<TrackCamera>();
                        EditorUtility.SetDirty(c);
                    }
                }
            }

            EditorGUILayout.Space(10);
            EditorGUILayout.LabelField("Image Source");

            // ImageSource DropDown
            int indexNew = EditorGUILayout.Popup("ImageSource", imageSourceDropdownIndex, imageSourceNames);
            if (indexNew != imageSourceDropdownIndex)
            {
                imageSource.enumValueIndex = indexNew;
                imageSourceDropdownIndex = indexNew;
                serializedObject.ApplyModifiedProperties();
            }

            // ImageSource specific UI
            if ((TrackManager.ImageSources)imageSource.enumValueIndex == TrackManager.ImageSources.Camera)
            {
                // Camera UI
                // Show info for Android
                if (EditorUserBuildSettings.activeBuildTarget == BuildTarget.Android)
                    EditorGUILayout.HelpBox("This USB camera ID only applies to the Editor.\nOn ANDROID the back-facing camera is used.", MessageType.Info, false);

                // PC camera ID
                EditorGUILayout.PropertyField(cameraID, new GUIContent("USB Camera ID"));
                // ID feedback
                if (cameraID.intValue < 0)
                    EditorGUILayout.HelpBox("The ID needs to be positive", MessageType.Error, false);
            }
            else
            {
                // FileSequence UI

                // Not on Android warning
                if (EditorUserBuildSettings.activeBuildTarget == BuildTarget.Android)
                    EditorGUILayout.HelpBox("The file sequence reader is not available on mobile platforms. \nOn ANDROID the back-facing camera is used.", MessageType.Warning, false);


                // Path input field
                EditorGUILayout.PropertyField(fileSquencePath, new GUIContent("File Sequence Path"));

                // Path feedback
                if(fileSquencePath.stringValue.Length == 0)
                    EditorGUILayout.HelpBox("Please provide a File Sequence Path", MessageType.Error, false);
                else
                {
                    if (!Directory.Exists(fileSquencePath.stringValue))
                        EditorGUILayout.HelpBox("The given path does not exist", MessageType.Error, false);
                    else if (Directory.GetFiles(fileSquencePath.stringValue).Length == 0)
                        EditorGUILayout.HelpBox("The given path is empty", MessageType.Error, false);
                    else
                        serializedObject.ApplyModifiedProperties();
                }

                // Calibration
                EditorGUILayout.PropertyField(fileSequenceCalibrationPath, new GUIContent("File Sequence Calibration Path"));
                // Path feedback
                if (fileSequenceCalibrationPath.stringValue.Length == 0)
                    EditorGUILayout.HelpBox("Please provide a calibration file for this specific sequence. Otherwise your USB camera calibration will be used. The image might not be displayed correctly.", MessageType.Warning, false);
                else
                    if (!File.Exists(fileSequenceCalibrationPath.stringValue))
                        EditorGUILayout.HelpBox("The given file does not exist", MessageType.Error, false);
                    else
                        if(Path.GetExtension(fileSequenceCalibrationPath.stringValue)!=".json")
                            EditorGUILayout.HelpBox("The given file is not a JSON file", MessageType.Error, false);
                        else
                            serializedObject.ApplyModifiedProperties();
            }

            GUILayout.EndVertical();
            // Apply values
            serializedObject.ApplyModifiedProperties();
        }

        /// <summary>
        /// Helper method to get all TrackManager used in a scene
        /// </summary>
        /// <param name="scene">Scene to search</param>
        /// <returns>List of used TrackManager in scene</returns>
        public static List<TrackManager> GetAllSceneManagers(Scene scene)
        {
            List<TrackManager> managers = new List<TrackManager>();
            TrackManager[] targetObjects = Resources.FindObjectsOfTypeAll<TrackManager>();
            foreach (TrackManager manager in targetObjects)
                if (manager.gameObject.scene == scene)
                    managers.Add(manager);

            return managers;
        }

        /// <summary>
        /// Verifies the correct configuration of the given manager
        /// </summary>
        /// <param name="manager">Manager script to check</param>
        /// <returns>Error string if error detected</returns>
        internal static string VerifyConfiguration(TrackManager manager)
        {
            // Ensure camera is attached
            Camera camera = manager.ARCamera;
            if (camera == null)
            {
                return "\"" + manager.gameObject.name + " \" has no scene camera attached";
            }

            // Ensure that referenced camera has VirnectCamera script attached
            TrackCamera cameraScript = camera.GetComponent<TrackCamera>();
            if (cameraScript == null)
            {
                return "\"" + manager.gameObject.name + " \" referenced camera has no \"TrackCamera\" component attached.";
            }

            //All checks passed
            return null;
        }

    }
}
