﻿// Copyright (C) 2020 VIRNECT CO., LTD.
// All rights reserved.
using UnityEngine;
using UnityEngine.Networking;
using System.IO;

namespace VIRNECT
{
#if UNITY_ANDROID
    public class AndroidTargetHelper
    {
        /// <summary>
        /// Files of a image target
        /// </summary>
        private static readonly string[] ImageTargetFiles = new string[4] { "TRACK_IMG_TARGET_CAM.cam", "TRACK_IMG_TARGET_IMG.dat", "TRACK_IMG_TARGET_KPT.dat", "TRACK_IMG_TARGET_CONFIG.json" };

        /// <summary>
        /// Files of a QR target
        /// </summary>
        private static readonly string[] QRTargetFiles = new string[2] { "TRACK_QRCODE_TARGET_CONFIG.json", "TRACK_QRCODE_TARGET_IMG.png" };

        /// <summary>
        /// Files of a CAD target
        /// </summary>
        private static readonly string[] CADTargetFiles = new string[2] { "TRACK_CAD_TARGET_CAD_MODEL.obj", "TRACK_CAD_TARGET_TEMPLATE.dat" };

        /// <summary>
        /// Files of a MAP target
        /// </summary>
        private static readonly string[] MAPTargetFiles = new string[2] { "TRACK_MAP_TARGET_CONFIG.json", "TRACK_MAP_TARGET_DATA.dat" };

        /// <summary>
        /// To overcome file access limitations of assets packet into the APK,
        /// target files are copied from inside the APK to the application specific
        /// persistentDataPath in the normal file system
        /// Attention: blocking method
        /// </summary>
        /// <param name="targets">Targets to copy</param>
        /// <returns>Path to access copied targets</returns>
        public static string ExtractTargets(string[] targets)
        {
            // The main directory to save the target files
            string resultPath = Application.persistentDataPath + Constants.targetRootDirectoryAndroid;

            // Process all targets
            foreach (string targetName in targets)
            {
                // Setup paths
                string folderName = Constants.targetPrefix + targetName;
                string sourcePath = "jar:file://" + Application.dataPath + Constants.targetAPKDirectoryAndroid + folderName;
                string goalPath = resultPath + folderName;

                // Copy target files if target root directory does not exist
                if (!Directory.Exists(goalPath))
                {
                    Directory.CreateDirectory(goalPath);

                    // Detect if image or QR target
                    string[] targetFiles = null;
                    switch (GetTargetType(sourcePath))
                    {
                        case TargetType.IMAGE: targetFiles = ImageTargetFiles; break;
                        case TargetType.QRCODE: targetFiles = QRTargetFiles; break;
                        case TargetType.CADMODEL: targetFiles = CADTargetFiles; break;
                        case TargetType.MAP: targetFiles = MAPTargetFiles; break;
                        default: Debug.LogError("Target type of \"" + targetName + "\" could not be determined"); return null;
                    }

                    // Copy all files
                    foreach (string fileName in targetFiles)
                    {
                        // Setup specific paths
                        string fromPath = sourcePath + "/" + fileName;
                        string toPath = goalPath + "/" + fileName;

                        // Use UnityWebRequest to access files inside APK
                        UnityWebRequest request = UnityWebRequest.Get(fromPath);
                        request.SendWebRequest();

                        // Wait until loaded
                        while (!request.isDone)
                            Debug.Log(request.downloadedBytes);

                        // Copy file if loaded correctly
                        if (request.isNetworkError || request.isHttpError)
                            Debug.Log(request.error);
                        else
                            File.WriteAllBytes(toPath, request.downloadHandler.data);
                    }
                }
            }

            //Prepare mapTargetFolder
            if (!Directory.Exists(Application.persistentDataPath + Constants.mapTargetRootDirectoryAndroid))
                Directory.CreateDirectory(Application.persistentDataPath + Constants.mapTargetRootDirectoryAndroid);

            return resultPath;
        }

        /// <summary>
        /// Determines the target type of the provided folder name by trying to load target specific files
        /// Attention: blocking method
        /// </summary>
        /// <param name="sourcePath">APK path</param>
        /// <returns>Target type of folder path. UNDEFINED if not determined</returns>
        private static TargetType GetTargetType(string sourcePath)
        {
            // Check image specific file:
            if (Exists(sourcePath + "/" + ImageTargetFiles[3]))
                return TargetType.IMAGE;

            // Check QR specific file:
            if (Exists(sourcePath + "/" + QRTargetFiles[0]))
                return TargetType.QRCODE;

            // Check CAD specific file:
            if (Exists(sourcePath + "/" + CADTargetFiles[0]))
                return TargetType.CADMODEL;

            // Check MAP specific file:
            if (Exists(sourcePath + "/" + MAPTargetFiles[1]))
                return TargetType.MAP;

            return TargetType.UNDEFINED;
        }

        /// <summary>
        /// Helper method to check if a file exists in the APK
        /// Attention: loads the file to check existence
        /// Attention: blocking method
        /// </summary>
        /// <param name="filePath">file to check</param>
        /// <returns>If the target exists</returns>
        private static bool Exists(string filePath)
        {
            UnityWebRequest request = UnityWebRequest.Get(filePath);
            request.SendWebRequest();
            while (!request.isDone) { }
            return !(request.isNetworkError || request.isHttpError);
        }
    }

#endif

}
